"""
Defines the data type of each argument for the circle tracking methods.
"""

circleTrackArgDTypes = {"radiusTolerance": int,
                     "offscreenParticles": bool,
                     "outlineOnly": bool,
                     "outlineThickness": float,
                     "negativeHalo": bool,
                     "haloThickness": float,
                     "negativeInside": bool,
                     "peakDownsample": int,
                     "minPeakPrevalence": float,
                     "intensitySoftmax": float,
                     "intensitySoftmin": float,
                     "invert": bool,
                     "allowOverlap": bool,
                     "fitPeaks": bool,
                     "debug": bool}

